
class Hash

  # Converts all keys in the Hash to be String values, returning a new Hash.
  # With a from_class parameter, limits conversion to only a certain class of keys.
  # It defaults to nil which convert any key class.
  #
  #   foo = { :name=>'Gavin', :wife=>:Lisa }
  #   foo.keys_to_s    #=>  { "name"=>"Gavin", "wife"=>:Lisa }
  #   foo.inspect      #=>  { :name =>"Gavin", :wife=>:Lisa }
  #
  #--
  # Adapted from Gavin Sinclair's #convert_keys.
  #++
  def keys_to_s( from_class=nil )
    raise ArgumentError, "Parameter must be a class" unless from_class.kind_of?(Class) if from_class
    self.dup.keys_to_s!( from_class )
  end

  # Aliases for better Extensions and Rails compatibility
  alias_method( :stringify_keys, :keys_to_s )

  # Synonym for Hash#keys_to_string, but modifies the receiver in place (and returns it).
  # With a from_class parameter, limits conversion to only a certain class of keys.
  # It defaults to nil which convert any key class.
  #
  #   foo = { :name=>'Gavin', :wife=>:Lisa }
  #   foo.keys_to_s!    #=>  { "name"=>"Gavin", "wife"=>:Lisa }
  #   foo.inspect       #=>  { "name"=>"Gavin", "wife"=>:Lisa }
  #
  #--
  # Adapted from Gavin Sinclair's #convert_keys.
  #++
  def keys_to_s!( from_class=nil )
    raise ArgumentError, "Parameter must be a class" unless from_class.kind_of?(Class) if from_class
    if from_class
      self.each_key{ |k|
        self[k.to_s]=self.delete(k) if k.respond_to?(:to_s) and k.class == from_class
      }
    else
      self.each_key{ |k|
        self[k.to_s]=self.delete(k) if k.respond_to?(:to_s)
      }
    end
    self
  end

  # Aliases for better Extensions and Rails compatibility
  alias_method( :stringify_keys!, :keys_to_s! )

end


#  _____         _
# |_   _|__  ___| |_
#   | |/ _ \/ __| __|
#   | |  __/\__ \ |_
#   |_|\___||___/\__|
#
=begin test

  require 'test/unit'

  class TCHash < Test::Unit::TestCase

    def test_keys_to_s
      foo = { :a=>1, :b=>2 }
      assert_equal( { "a"=>1, "b"=>2 }, foo.keys_to_s )
      assert_equal( { :a =>1, :b=>2 }, foo  )
    end

    def test_keys_to_s!
      foo = { :a=>1, :b=>2 }
      assert_equal( { "a"=>1, "b"=>2 }, foo.keys_to_s!  )
      assert_equal( { "a"=>1, "b"=>2 }, foo )
    end

  end

=end
