# = taskable.rb
#
# == Copyright (c) 2006 Thomas Sawyer
#
#   Ruby License
#
#   This module is free software. You may use, modify, and/or redistribute this
#   software under the same terms as Ruby.
#
#   This program is distributed in the hope that it will be useful, but WITHOUT
#   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#   FOR A PARTICULAR PURPOSE.
#
# == Auhtor(s)
#
# * Thomas Sawyer

# Author::    Thomas Sawyer
# Copyright:: Copyright (c) 2006 Thomas Sawyer
# License::   Ruby License

require 'tsort'
require 'facets/core/kernel/cache.rb'
require 'facets/core/kernel/method.rb'
require 'facets/more/inheritor.rb'

# = TSort
#
# TSort is used to determine which prerequisites to run
# for a task.
#
# Here we add a convenience method --if such a long method
# name can be said to be convenient ;) Although we ended
# not using it after all --it still may proved useful and
# may eventually be added to Facets or submitted as a patch.

module TSort
  def strongly_connected_components_from(node)
    result = []
    each_strongly_connected_component_from(node) do |component|
      result << component
    end
    result
  end
end

# = Task
#
# A Task is a Method with prerequisite dependencies.
#
# NOTE Can't subclass Method b/c Method has no initiator.
# Not sure that would be a good idea anyway.

class Task
  attr_accessor :description

  def initialize( name, container, preq, desc=nil, &action )
    @name = name
    @container = container
    @prerequisite = preq || []
    @description = desc
    @action = action
  end

  def prerequisite
    @prerequisite ||= []
  end

#   #--
#   # TODO Problem is 'a' can't take any parameters,
#   #      and there's no way to sort the tasks
#   #      from here. FIX if possible.
#   #++
#   def to_proc
#     r = @prerequisite
#     n = @name
#     a = @action
#     lambda {
#       r.each{ |d| send(d) unless n == d }
#       instance_eval &a if a
#     }
#   end

  def complete
    # ensure no cycles in the graph and extract call graph
    #own = (class << self; self; end) # in case there are singletons?
    t = Graph.new( container, name )
    t.each_strongly_connected_component_from( name ) { |c|
      #@container.instance_task(c).call
      @container.send("->#{d}")
    }
  end

  def call
    @action.call
  end

  def each(&blk)
    @prerequisite.each(&blk)
  end

  # Task::Graph is a Tarjan-sorted hash.
  # It is used to generate the proper dependency chains.
  # TODO Make Multiton off of base.

  class Graph < Hash
    include TSort

    def initialize( base, name )
      @base = base.class
      @name = name
      instance_map( name )
    end

    def instance_map( name )
      t = @base.instance_task(name)
      raise "prerequired task does not exist -- #{name} in #{@base}" unless t
      self[name] = t.prerequisite || []
      self[name].each { |r| instance_map( r ) }
    end

    alias_method :tsort_each_node, :each_key

    def tsort_each_child(node, &block)
      fetch(node).each(&block)
    end
  end

end

# = Taskable
#
# The Taskable module provides a generic task system
# patterned after Rake, but useable in any
# code context --not just with the Rake tool. In other
# words one can create methods with dependencies.
#
# NOTE Unlike methods, tasks can't take parameters
# if they are to be used as prerequisites.
# This capability might be added in the future, but
# presently syntax limitations make it awkward.

module Taskable

  # Provide a mapping of tasks, including task from ancestors.
  # TODO Not sure this is properly reading from ancestors.

  # Define description for subsequent task.

  def desc(line=nil)
    return @last_description unless line
    @last_description = line.gsub("\n",'')
  end

  # Use a description for subsequent task.

  def desc!
    l = @last_description
    @last_description = nil
    l
  end

  # Access a task.

  def instance_task( name )
    instance_task_table[ name.to_sym ]
  end

  # List of task names.

  def instance_tasks
    instance_task_table.keys
  #  t = instance_methods(true).select{ |m| m =~ /^->/ }
  #  t.collect { |e| e[2..-1].to_sym }
  end

  # List of task names with descriptions.

  def described_tasks
    instance_tasks.select { |name| instance_task(name).description }
  end

  # List of task names without descriptions.

  def undescribed_tasks
    instance_tasks.select { |name| ! instance_task(name).description }
  end

  # Define a task.

  def task(args, &action)
    if Hash === args
      raise ArgumentError, "#{args.size} for 1" if args.size != 1
      name, *deps = *(args.to_a.flatten)
      name = name.to_sym
      deps = deps.compact.collect{ |e| e.to_sym }
    else
      name, deps = args.to_sym, []
    end
    # create task via create core and callable methods.
    # We do it this way b/c otherwise we'd have to use
    # instance_eval and then can't ever pass paramaters.
    if action
      define_method( "->#{name}", &action ) # core of the task
      define_method( name ) do
        # ensure no cycles in the graph and extract call graph
        #own = (class << self; self; end) # in case there are singletons?
        todolist = Task::Graph.new( self, name )
        todolist.each_strongly_connected_component_from( name ) do |c|
          send("->#{c}")
        end
      end
      instance_task_table![name] = Task.new( name, self, deps, desc!, &action )
    elsif !instance_task_table![name]
      #define_method( "->#{name}", lambda{} )  # empty action
      define_method( name ) do
        # ensure no cycles in the graph and extract call graph
        #own = (class << self; self; end) # in case there are singletons?
        todolist = Task::Graph.new( self, name )
        todolist.each_strongly_connected_component_from( name ) do |c|
          send("->#{c}") unless "#{c}" == "#{name}"
        end
      end
      instance_task_table![name] = Task.new( name, self, deps, desc!, &action )
    else
      m = instance_task_table![name]
      #m.description = desc!  # TODO should this apply only if there is an action?
      m.prerequisite.concat deps unless deps.empty?
      m
    end
  end

  # Include the Task module into the Module class.

  class ::Module
    include Taskable
    # This has to be down here and not in Taskable b/c of
    # a glitch in inheritor's design.
    inheritor :instance_task_table, {}, :merge
  end

end

#

class Object
  def send_task( name )
    (class << self; self; end).instance_task( name )
  end
end

#--
# TODO Allows tasks to defined at the toplevel (main).
# This may need work b/c it might cause an infinite loop.
# Plus there are a few method missing yet. Anyway,
# I'm not needing this functionality at the moment, and
# I leave it to somone else to finish when they do.
#
# class Object
#
#   def instance_tasks
#     (class << self; self; end).instance_tasks
#   end
#
#   def instance_task(*args, &blk)
#     (class << self; self; end).task(*args, &blk)
#   end
#
#   def desc(*args, &blk)
#     (class << self; self; end).desc(*args, &blk)
#   end
#
#   def desc!(*args, &blk)
#     (class << self; self; end).desc!(*args, &blk)
#   end
#
#   def task(*args, &blk)
#     (class << self; self; end).desc(*args, &blk)
#   end
#
# end
#++


#  _____         _
# |_   _|__  ___| |_
#   | |/ _ \/ __| __|
#   | |  __/\__ \ |_
#   |_|\___||___/\__|
#
=begin test

  require 'test/unit'

  class TCTask < Test::Unit::TestCase

    module M
      task :m1 => [ :c1 ] do @x << "m1" end
      task :m2 => [ :c2 ] do @x << "m2" end
      task :m3 => [ :c3 ] do @x << "m3" end
      task :m4 => [ :c1, :c2, :c3 ] do @x << "m4" end
      task :m5 do @x << 'm5' end
    end

    class B
      attr :x
      def initialize ; @x = [] ; end

      task :b1 do @x << "b1" end
      task :b2 => [ :b1 ]
    end

    class C
      include M
      attr :x
      def initialize ; @x = [] ; end

      task :c1 do @x << "c1" end
      task :c2 => [ :c1, :c1 ] do @x << "c2" end
      task :c3 => [ :c1, :c2 ] do @x << "c3" end
      task :c4 => [ :m1 ] do @x << "c4" end
      task :c5 => [ :c5 ] do @x << "c5" end
      task :c6 => [ :c7 ] do @x << "c6" end
      task :c7 => [ :c6 ] do @x << "c7" end
    end

    class D < C
      task :d1 => [ :c1 ] do @x << "d1" ; end
      task :d2 => [ :m1 ] do @x << "d2" ; end
    end

    module N
      include M
    end

    class E
      include N
      attr :x
      def initialize ; @x = [] ; end

      task :e1 => [ :c1 ] do @x << "e1" ; end
      task :e2 => [ :m1 ] do @x << "e2" ; end
      task :e3 => [ :m5 ] do @x << "e3" ; end
    end

    module O
      attr :x
      task :o1 do (@x||=[]) << "o1" end
      task :o2 => [ :o1 ] do (@x||=[]) << "o2" end
    end

    # TODO Appearently extending does play well with inheritor?
    # F = Module.new { extend O }

    # tests

    def test_B1
      b = B.new ; b.b1
      assert_equal( [ 'b1' ], b.x )
    end

    def test_B2
      b = B.new ; b.b2
      assert_equal( [ 'b1' ], b.x )
    end

    def test_C1
      c = C.new ; c.c1
      assert_equal( [ 'c1' ], c.x )
    end

    def test_C2
      c = C.new ; c.c2
      assert_equal( [ 'c1', 'c2' ], c.x )
    end

    def test_C3
      c = C.new ; c.c3
      assert_equal( [ 'c1', 'c2', 'c3' ], c.x )
    end

    def test_C4
      c = C.new ; c.c4
      assert_equal( [ 'c1', 'm1', 'c4' ], c.x )
    end

    def test_M1
      c = C.new ; c.m1
      assert_equal( [ 'c1', 'm1' ], c.x )
    end

    def test_M2
      c = C.new ; c.m2
      assert_equal( [ 'c1', 'c2', 'm2' ], c.x )
    end

    def test_M3
      c = C.new ; c.m3
      assert_equal( [ 'c1', 'c2', 'c3', 'm3' ], c.x )
    end

    def test_M4
      c = C.new ; c.m4
      assert_equal( [ 'c1', 'c2', 'c3', 'm4' ], c.x )
    end

    def test_D1
      d = D.new ; d.d1
      assert_equal( [ 'c1', 'd1' ], d.x )
    end

    def test_D2
      d = D.new ; d.d2
      assert_equal( [ 'c1', 'm1', 'd2' ], d.x )
    end

    def test_E1
      e = E.new
      assert_raises( RuntimeError ) { e.e1 }
      #assert_equal( [ 'c1', 'e1' ], e.x )
    end

    def test_E2
      e = E.new
      assert_raises( RuntimeError ) { e.e2 }
      #assert_equal( [ 'c1', 'm1', 'e2' ], e.x )
    end

    def test_E3
      e = E.new ; e.e3
      assert_equal( [ 'm5', 'e3' ], e.x )
    end

    # def test_F1
    #   F.o1
    #   assert_equal( [ 'o1' ], F.x )
    # end
    #
    # def test_F2
    #   F.o2
    #   assert_equal( [ 'o1', 'o1', 'o2' ], F.x )
    # end

  end

=end
