#--
# GLogger
#
# Copyright (c) 2005 George Moschovitis <gm@navel.gr>
#
# Ruby License
#
# This module is free software. You may use, modify, and/or redistribute this
# software under the same terms as Ruby.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.
#
# ===========================================================================
#  Revision History ::
# ---------------------------------------------------------------------------
#
#  2005.08.25  trans   Ported to Mega Modules with Mark Probet's changes.
#
# ===========================================================================
#
# :TODO: Docs upadte, What compatibility purposes?
#
# :TODO: Implement as mixin that can be added to Log4R.
#
#++

require 'logger'

#:title: GLogger
#
# A simple wrapper arround the Ruby logger. Mainly for
# compatibility purposes.
#
# == Usage
#
#  TODO
#
# Author(s)
#
# * George Moschovitis
# * Gavin Sinclair
# * Mark Probet
#

class GLogger < Logger
  alias_method :devel, :debug
  alias_method :fine, :debug

  # Prints a trace message to DEBUGLOG (at debug level).  
  # Useful for emitting the value of variables, etc.  Use 
  # like this:
  #
  #   x = y = 5
  #   trace 'x'        # -> 'x = 5'
  #   trace 'x ** y'   # -> 'x ** y = 3125'
  #
  # If you have a more complicated value, like an array of 
  # hashes, then you'll probably want to use an alternative 
  # output format.  For instance:
  #
  #   trace 'value', :yaml
  #
  # Valid output format values (the _style_ parameter) are:
  #
  #   :p :inspect
  #   :pp                     (pretty-print, using 'pp' library)
  #   :s :to_s
  #   :y :yaml :to_yaml       (using the 'yaml' library')
  #
  # The default is <tt>:p</tt>.
  #
  # CREDITS:
  #
  # This code comes straight from the dev-utils Gem.
  # Author: Gavin Sinclair <gsinclair@soyabean.com.au>

  def trace(expr, style=:p)
    unless expr.respond_to? :to_str
      warn "trace: Can't evaluate the given value: #{caller.first}"
    else
      require 'nano/binding/of_caller'

      Binding.of_caller do |b|
        value = b.eval(expr.to_str)
        formatter = TRACE_STYLES[style] || :inspect
        case formatter
        when :pp then require 'pp'
        when :y, :yaml, :to_yaml then require 'yaml'
        end
        value_s = value.send(formatter)
        message = "#{expr} = #{value_s}"
        lines = message.split(/\n/)
        indent = "   "
        debug(lines.shift)                      
        lines.each do |line|
          debug(indent + line)
        end
      end
    end
  end

  TRACE_STYLES = {}  # :nodoc:
  TRACE_STYLES.update(
    :pp => :pp_s, :s => :to_s, :p => :inspect,
    :y => :to_yaml, :yaml => :to_yaml,
    :inspect => :inspect, :to_yaml => :to_yaml
  )

  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  # Global logger interface. 
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  # Override the logger with your custom version.

  @@global_logger = GLogger.new(STDERR)

  # Set the global Logger.

  def self.set(logger)
    if logger.is_a?(String)
      @@global_logger = GLogger.new(logger)
    elsif logger.is_a?(Logger)
      @@global_logger = logger
    else
      raise ArgumentError.new
    end
  end

  def self.get
    @@global_logger
  end

  def self.warn(str)
    @@global_logger.warn(str)
  end

  def self.info(str)
    @@global_logger.info(str)
  end

  def self.debug(str)
    @@global_logger.debug(str)
  end

  def self.error(str)
    @@global_logger.error(str)
  end

  #--
  # Saddly have to duplicate the code to make
  # Binding.of_caller work.
  #++
  def self.trace(expr, style=:p)
    unless expr.respond_to? :to_str
      warn "trace: Can't evaluate the given value: #{caller.first}"
    else
      require 'nano/binding/of_caller'

      Binding.of_caller do |b|
        value = eval(expr.to_str, b)
        formatter = TRACE_STYLES[style] || :inspect
        case formatter
        when :pp then require 'pp'
        when :y, :yaml, :to_yaml then require 'yaml'
        end
        value_s = value.send(formatter)
        message = "#{expr} = #{value_s}"
        lines = message.split(/\n/)
        indent = "   "
        debug(lines.shift)                      
        lines.each do |line|
          debug(indent + line)
        end
      end
    end
  end

  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  private

  # the default Ruby logger has a hardwired silly format.

  Format = "%5s: %s\n"

  def format_message(severity, timestamp, msg, progname)
    Format % [severity, msg]
  end

end

# Add logging capabilities to the including class.
# When using the log/logger variables always check
# for nil:
#
#   log.info "your #{expensive_calculation} comes here" if @log
#
# This way the expensive calculation is avoided if the
# logger is dissabled (@log == nil).
#
module Logging
  attr_accessor :logger
  alias_method :log, :logger
end

